//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
// 
// Permission is hereby granted, free of charge, to any person obtaining a 
// copy of this software and associated documentation files (the "Software"), 
// to deal in the Software without restriction, including without limitation 
// the rights to use, copy, modify, merge, publish, distribute, sublicense, 
// and/or sell copies of the Software, and to permit persons to whom the 
// Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included 
// in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES 
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
// OTHER DEALINGS IN THE SOFTWARE.
// 
// Except as contained in this notice, the name of Dallas Semiconductor 
// shall not be used except as stated in the Dallas Semiconductor 
// Branding Policy. 
//---------------------------------------------------------------------------
//
//  ibshadeb.c - This utility uses TMEX to debit one unit from a SHA iButton 
//               (DS1963S).
//
//  Compiler:  Microsoft Visual C 
//  Version:   1.01
//  History:   
//             1.01 -> 1.02  Used correct MUC (Monetary Units Code) and 
//                           multiplier for US dollars

#define TMEXUTIL

#include "ibtmexcw.h"   
#include "ibsha.h"
#include "uline.h"

// mode constants
enum { MD_FIND_ROVING=0, MD_READ_AUTH, MD_MONEY_VERF, MD_MONEY_CHANGE,
       MD_MONEY_WRITE, MD_READ_BACK_AUTH, MD_VERIFY_GONE };

// external function prototypes 
extern int RunScriptSHA(Script script[], void *, int);
extern void ExitProg(char *, int);
extern int EnterNum(char *, int, long *, long, long);

// local function prototypes
int PrintHeader(char *);
void output_status(int, char *);
long SetNewMoneyValue(eCertificate *,long);
int DebitRovingSHAiButton(int,long);
void eCertPrint(eCertificate *, char *);
int SetMaxBlock(void);

// globals 
ulong time_stamp;

//----------------------------------------------------------------------
//  This is the Main routine for ibshadeb 
//
void main(short argc, char **argv)
{
   int ShowStatus=FALSE,filenum,LOOP,i,cnt,exit_code=EXIT_SUCCESS;
   char msg[200];
   int amount;

   // check arguments to see if request instruction with '?' or too many
   if ((argc > 4) || ((argc > 1) && (argv[1][0] == '?' || argv[1][1] == '?')))
       ExitProg("\nusage: ibshadeb <output_filename> </Verbose> </Loop>\n"
              "  - debit from a roving SHA iButton using coprocessor SHA iButton\n"
              "    on same port. (use SetPrt32.exe to set default)\n"
              "  - <output_filename> optional output filename\n"
              "  - </V> optional show verbose output\n"
              "  - </L> optional continuous debit loop\n"
              "  - press any key to stop\n"
              "  - version 1.02\n",EXIT_ERROR);

   // read the default PortNum and PortType 
   if (!TMReadDefaultPort(&PortNum[IB_ROVING],&PortType[IB_ROVING]))
      ExitProg("ERROR, Could not read the default PortNum and PortType!\n",EXIT_ERROR);

   // for now hard code CO_PROCESSOR port to be one more
   PortNum[IB_COPR] = PortNum[IB_ROVING];
   PortType[IB_COPR] = PortType[IB_ROVING];

   VERBOSE = FALSE;
   LOOP = FALSE;

   // print the header info 
   if (!PrintHeader("ibshadeb Version 1.02"))
      exit(1);

   // check arguments 
   filenum = 0;
   for (i = 1; i < argc; i++)
   {
      if (argv[i][0] != '/')
         filenum = i;
      else if ((argv[i][1] == 'V') || (argv[i][1] == 'v'))
         VERBOSE = TRUE;
      else if ((argv[i][1] == 'l') || (argv[i][1] == 'L'))
         LOOP = TRUE;
   }

   // open the output file  
   fp = NULL;
   if (filenum > 0)
   {
      fp = fopen(argv[filenum],"wb");
      if(fp == NULL)
      {    
         printf("ERROR, Could not open output file!\n");
         exit(1);
      }
      else
         printf("File '%s' opened to write sensor data.\n\n",
                 argv[filenum]);
   }

   // run the script to setup the 1-Wire
   if (RunScriptSHA(Verify1Wire, NULL, 0))
      output_status(LV_ALWAYS,"** 1-Wire port verified\n\n");
   else
      ExitProg("1-Wire Net not successfuly setup, end program\n",EXIT_ERROR);

   // enter amount to debit
   amount = 50;
   if (!EnterNum("\nEnter the amount to deduct per transaction (in cents)\n", 7, &amount, 0, 9999999))
      ExitProg("Aborted entering, end program\n",EXIT_ERROR);
   else
   {
      sprintf(msg,"Amount to deduct: $%4.2f US\n\n",(float)(amount/100.0));
      output_status(LV_ALWAYS,msg);  
   }

   // attempt to open session on 1-Wire net
   if (RunScriptSHA(OpenSession, NULL, 0))
   {
      output_status(LV_VERBOSE,"** Session on 1-Wire opened\n");

      // attempt to find co-processor
      provider[0] = 0;
      if (RunScriptSHA(FindCoprocessor, &copr_rom[0], sizeof(copr_rom)))
      {      
         // see if provider has been set (real result of FindCopr)
         if (provider[0])
         {
            // loop to print ROM of device found
            cnt = sprintf(msg,"** Co-processor SHA iButton found: ");
            for (i = 7; i >= 0; i--)
               cnt += sprintf(msg + cnt,"%02X",copr_rom[i]);
            cnt += sprintf(msg + cnt,"\n** Provider file: ");
            for (i = 0; i < 4; i++)
               cnt += sprintf(msg + cnt,"%c",provider[i]);
            cnt += sprintf(msg + cnt,".%d\n",provider[4]);
            cnt += sprintf(msg + cnt,"** Money master secret page: %d\n",c_mmaster_scrt);
            cnt += sprintf(msg + cnt,"** Auth master secret page: %d\n",c_amaster_scrt);
            cnt += sprintf(msg + cnt,"** Unique device (calculated) secret page: %d\n\n",c_udevice_scrt);
            output_status(LV_ALWAYS,msg);
            
            // find and debit roving SHA iButtons
            exit_code = DebitRovingSHAiButton(LOOP,-amount);

            // absorb any key presses
            while (kbhit())
               getch();
         }
         else
         {
            output_status(LV_ALWAYS,"ERROR, could not find co-processor on 1-Wire Net, end program\n");
            exit_code = EXIT_ERROR;
         }
      }
      else
      {
         output_status(LV_ALWAYS,"ERROR, could not find co-processor on 1-Wire Net, end program\n");
         exit_code = EXIT_ERROR;
      }

      // close session
      if (RunScriptSHA(CloseSession, NULL, 0))
         output_status(LV_VERBOSE,"** Session on 1-Wire closed\n");
   }
   else
   {
      output_status(LV_ALWAYS,"ERROR, Could not open co-processor 1-Wire port\n");
      exit_code = EXIT_ERROR;
   }

   // close opened file
   if ((fp != NULL) && (fp != stdout))
   {
      printf("File '%s' closed.\n",
              argv[filenum]);
      fclose(fp);
   }

   // check on result of operation
   if (exit_code)
      printf("ERROR reported (%d), end program\n",exit_code);
   else
      printf("End program normally\n");

   // end the program
   exit(exit_code);
}


//--------------------------------------------------------------------------
// Set new money amount
//
long SetNewMoneyValue(eCertificate *ec, long amount_change)
{
   long l,money;
   int i;

   // convert balance to ulong
   l = 0;
   for (i = 2; i >= 0; i--)   
   {
      l <<= 8;
      l |= ec->balance[i];
   }

   // change by 'amount_change'
   if (l + amount_change < 0)
      l = 0;
   else
      l += amount_change;
   money = l;

   // put back in balance
   for (i = 0; i < 3; i++)
   {
      ec->balance[i] = (uchar)(l & 0xFF);
      l >>= 8;
   }

   // set transaction ID to mine
   ec->trans_id = 0x1234;

   // also increment counter
   ec->write_cycle_cnt++;

   return money;
}


//--------------------------------------------------------------------------
// Print the eCertificate
//
void eCertPrint(eCertificate *ec, char *msg)
{
   ulong ul;
   int i,mcnt=0;

   // device time to transact
   time_stamp = GetTickCount() - time_stamp;

   mcnt = sprintf(msg,"---------------------------\n");   
   mcnt += sprintf(msg + mcnt,"| Device: ");
 
   for (i = 6; i >=0; i--)
      mcnt += sprintf(msg + mcnt,"%02X",ec->serial_num[i]);
  
   mcnt += sprintf(msg + mcnt,"\n| User data: %s\n",ec->user_data);

   // convert balance to ulong
   ul = 0;
   for (i = 2; i >= 0; i--)   
   {
      ul <<= 8;
      ul |= ec->balance[i];
   }

   mcnt += sprintf(msg + mcnt,"| Balance: $%4.2f US\n",(float)(ul/100.0)); // (1.02)
   mcnt += sprintf(msg + mcnt,"| Transaction time: %d ms\n",time_stamp);
   mcnt += sprintf(msg + mcnt,"---------------------------\n");   
}


//--------------------------------------------------------------------------
// loop to find and debit sha iButtons
//
int DebitRovingSHAiButton(int LOOP, long amount)
{
   eCertificate ec,vec;
   short ROM[8];
   int debit_mode,check_present,mcnt;
   char msg[255],lastmsg[255];
   ulong ts;

   // start mode
   debit_mode = MD_FIND_ROVING;

   // loop to find and deduct from the money account on roving
   do
   {
      // assume need to check if device is present
      check_present = TRUE;

      // switch on current debit mode
      switch (debit_mode)
      {
         case MD_FIND_ROVING:
            // time stamp of start of search
            time_stamp = GetTickCount(); 
            if (RunScriptSHA(FindRoving, &ROM[0], sizeof(ROM)))
            {
               sprintf(msg,"** Roving SHA iButton found\n");
               debit_mode = MD_READ_AUTH;
            }
            else  
               sprintf(msg,"\n** Waiting for roving SHA iButton\n");
            check_present = FALSE;
            break;

         case MD_READ_AUTH:
            if (RunScriptSHA(ReadAuth, &ec, sizeof(ec)))
            {
               sprintf(msg,"** Read authentication complete\n");
               check_present = FALSE;
               debit_mode = MD_MONEY_VERF;
            }
            else  
               sprintf(msg,"ERROR doing read authentication\n");
            break;

         case MD_MONEY_VERF:
            // first check if correct money type (1.02)
            if (ec.code_mult != 0x8B48)
            {
               sprintf(msg,"\n** ERROR, money wrong type (code %04X)\n\n",ec.code_mult);
               check_present = FALSE;
               debit_mode = MD_VERIFY_GONE;
               break;     
            }
            // verify money valid
            if (RunScriptSHA(VerifyMoney, &ec, sizeof(ec)))
            {
               sprintf(msg,"** Money value authenticated\n");
               check_present = FALSE;
               debit_mode = MD_MONEY_CHANGE;
            }
            else  
               sprintf(msg,"ERROR authenticating money\n");
            break;

         case MD_MONEY_CHANGE:
            // new value for the money
            sprintf(msg,"** New money value set (debit of %d)\n",-amount);
            SetNewMoneyValue(&ec,amount);
            check_present = FALSE;
            debit_mode = MD_MONEY_WRITE;
            break;

         case MD_MONEY_WRITE:
            if (RunScriptSHA(WriteMoney, &ec, sizeof(ec)))
            {
               sprintf(msg,"** New money written\n");
               check_present = FALSE;
            }
            else  
               sprintf(msg,"ERROR writing new money value\n");
            debit_mode = MD_READ_BACK_AUTH;
            break;

         case MD_READ_BACK_AUTH:
            if (RunScriptSHA(ReadBackAuth, &vec, sizeof(vec)))
            {
               check_present = FALSE;
               mcnt = sprintf(msg,"** Read back authentication complete\n");
   
               // check if written correclty
               if (ec.write_cycle_cnt == vec.write_cycle_cnt)
               {
                  // succes so print result
                  eCertPrint(&ec,&msg[mcnt]);
                  // make sure max block ok
                  SetMaxBlock();
                  if (LOOP)
                  {
                     // if in loop mode go directly to read auth
                     // (skip finding roving)
                     debit_mode = MD_READ_AUTH; 
                     // start the time stamp now
                     time_stamp = GetTickCount(); 
                  }
                  else
                     debit_mode = MD_VERIFY_GONE;
               }   
               // not written so go back to that mode
               else if (ec.write_cycle_cnt == (vec.write_cycle_cnt + 1))
                  debit_mode = MD_MONEY_WRITE;
               // unknown error value
               else
                  return EXIT_ERROR;
            }
            else  
               sprintf(msg,"** Error doing read back authentication\n");
            break;

         case MD_VERIFY_GONE:
            check_present = FALSE;
            if (!RunScriptSHA(RovingIsPresent, NULL, 0))
            {
               debit_mode = MD_FIND_ROVING;
               sprintf(msg,"\n** Waiting for roving SHA iButton\n");
            }
            else  
               sprintf(msg,"** Waiting for roving SHA iButton to leave\n");
            break;

         default:
            debit_mode = MD_FIND_ROVING;
            break;

      };

      // error occured so check if roving device present
      if (check_present && !kbhit())
      {
         // loop for up to 500ms waiting for device to return
         ts = GetTickCount() + 500;
         do {} 
         while (!RunScriptSHA(RovingIsPresent, NULL, 0) && (ts > GetTickCount()));
         // if not return then print message
         if (!RunScriptSHA(RovingIsPresent, NULL, 0))
         {
            sprintf(msg,"\n>>>>>>>>>>   Please present roving iButton to finish transaction\n\n");
            MessageBeep(MB_OK);
         }
      }

      // send output
      if (strcmp(lastmsg,msg) != 0)
      {
         output_status(LV_ALWAYS,msg);
         sprintf(lastmsg,"%s",msg);
      }
      else
         output_status(LV_VERBOSE,".");
   }
   while (!kbhit());

   return EXIT_SUCCESS;
}

//--------------------------------------------------------------------------
// Set max block size to 16
//
int SetMaxBlock(void)
{
   LineType Line;
   char buf[200];
   static int cnt=0;

   // read line settings first (so you only change the one you want)
   sprintf(Line.ID,"DS2480");
   Line.Operation = READ_FUNCTION;
   if (TMLine(session_handle[dev],&Line) == 1)
   {
      // change only if not 16
      if (Line.MAXSEND != 16)
      {
         sprintf(buf,"######### setting back to 16 max block ####### (%d)\n",cnt++); 
         output_status(LV_VERBOSE,buf);

         // change max block size to 16 bytes
         Line.MAXSEND = 16;

         // now write all of the values back
         Line.Operation = WRITE_FUNCTION;
         if (TMLine(session_handle[dev],&Line) != 1)
             // error
             return FALSE;
      }
   }
   else
      // error
      return FALSE;

   return TRUE;
}

